Document: tolua++
Title: Debian tolua++ Manual
Author: <insert document author here>
Abstract: This manual describes what tolua++ is
 and how it can be used to
 manage online manuals on Debian systems.
Section: unknown

Format: debiandoc-sgml
Files: /usr/share/doc/tolua++/tolua++.sgml.gz

Format: postscript
Files: /usr/share/doc/tolua++/tolua++.ps.gz

Format: text
Files: /usr/share/doc/tolua++/tolua++.text.gz

Format: HTML
Index: /usr/share/doc/tolua++/html/index.html
Files: /usr/share/doc/tolua++/html/*.html

  
